"""coBib's UIs.

coBib is designed with two user interfaces in mind:
- a command-line interface (short: CLI)
- a terminal user interface (short: TUI)
"""

from .cli import CLI as CLI
from .tui import TUI as TUI
from .ui import UI as UI
